from . import models
from huscy.project_design.models import DataAcquisitionMethod, Experiment, Session


def create_experiment(project, title='', description=''):
    order = project.experiments.count()

    return Experiment.objects.create(
        description=description,
        order=order,
        project=project,
        title=title or f'Experiment {order + 1}',
    )


def create_session(experiment, duration, title=''):
    order = experiment.sessions.count()

    return Session.objects.create(
        duration=duration,
        experiment=experiment,
        order=order,
        title=title or f'Session {order + 1}',
    )


def create_data_acquisition_method(session, type, location, stimulus=None):
    order = session.data_acquisition_methods.count() + 1

    if isinstance(type, models.DataAcquisitionMethodType):
        pass
    elif isinstance(type, str):
        type = models.DataAcquisitionMethodType.objects.get(pk=type)
    else:
        raise ValueError('Unknown data type for `type` attribute')

    return DataAcquisitionMethod.objects.create(
        location=location,
        order=order,
        session=session,
        stimulus=stimulus,
        type=type,
    )


def get_experiments(project):
    return project.experiments.order_by('order')


def get_sessions(experiment):
    return experiment.sessions.order_by('order')


def get_data_acquisition_methods(session):
    return session.data_acquisition_methods.order_by('order')


def get_data_acquisition_method_type(short_name):
    return models.DataAcquisitionMethodType.objects.get(short_name=short_name)


def update_experiment(experiment, **kwargs):
    updatable_fields = (
        'description',
        'title',
    )
    return update(experiment, updatable_fields, **kwargs)


def update_session(session, **kwargs):
    updatable_fields = (
        'duration',
        'setup_time',
        'teardown_time',
        'title',
    )
    return update(session, updatable_fields, **kwargs)


def update_data_acquisition_method(data_acquisition_method, **kwargs):
    updatable_fields = (
        'stimulus',
        'location',
    )
    return update(data_acquisition_method, updatable_fields, **kwargs)


def update(instance, updatable_fields, **kwargs):
    update_fields = []

    for field_name, value in kwargs.items():
        if field_name not in updatable_fields:
            raise ValueError(f'Cannot update field "{field_name}".')
        setattr(instance, field_name, value)
        update_fields.append(field_name)

    if update_fields:
        instance.save(update_fields=update_fields)

    return instance



from datetime import timedelta

from django.db import models
from django.utils.translation import gettext_lazy as _

from huscy.projects.models import Project


class Experiment(models.Model):
    project = models.ForeignKey(Project, on_delete=models.CASCADE, related_name='experiments',
                                verbose_name=_('Project'))

    title = models.CharField(_('Title'), max_length=255)
    description = models.TextField(_('Description'), blank=True, default='')

    order = models.PositiveSmallIntegerField(_('Order'))

    class Meta:
        verbose_name = _('Experiment')
        verbose_name_plural = _('Experiments')


class Session(models.Model):
    experiment = models.ForeignKey(Experiment, on_delete=models.CASCADE, related_name='sessions',
                                   verbose_name=_('Experiment'))

    title = models.CharField(_('Title'), max_length=255)

    order = models.PositiveSmallIntegerField(_('Order'))

    setup_time = models.DurationField(_('Setup time'), default=timedelta())
    duration = models.DurationField(_('Duration'))
    teardown_time = models.DurationField(_('Teardown time'), default=timedelta())

    class Meta:
        verbose_name = _('Session')
        verbose_name_plural = _('Sessions')


class DataAcquisitionMethodType(models.Model):
    short_name = models.CharField(_('Short name'), max_length=32, primary_key=True, editable=False)
    name = models.CharField(_('Name'), max_length=255)

    class Meta:
        verbose_name = _('Data acquisition method type')
        verbose_name_plural = _('Data acquisition method types')


class DataAcquisitionMethod(models.Model):
    class STIMULUS(models.TextChoices):
        auditive = ('auditive', _('Auditive'))
        gustatory = ('gustatory', _('Gustatory'))
        haptic = ('haptic', _('Haptic'))
        olfactory = ('olfactory', _('Olfactory'))
        visual = ('visual', _('Visual'))

    session = models.ForeignKey(Session, on_delete=models.CASCADE,
                                related_name='data_acquisition_methods', verbose_name=_('Session'))
    order = models.PositiveSmallIntegerField(_('Order'))

    stimulus = models.CharField(_('Stimulus'), max_length=32, null=True, choices=STIMULUS.choices)

    type = models.ForeignKey(DataAcquisitionMethodType, on_delete=models.PROTECT,
                             verbose_name=_('Type'))

    location = models.CharField(_('Location'), max_length=126)

    class Meta:
        verbose_name = _('Data acquisition method')
        verbose_name_plural = _('Data acquisition methods')



from django.urls import include, path
from rest_framework_nested.routers import NestedDefaultRouter

from huscy.project_design import views
from huscy.projects.urls import project_router


project_router.register('experiments', views.ExperimentViewSet, basename='experiment')

experiment_router = NestedDefaultRouter(project_router, 'experiments', lookup='experiment')
experiment_router.register('sessions', views.SessionViewSet, basename='session')

session_router = NestedDefaultRouter(experiment_router, 'sessions', lookup='session')
session_router.register('dataacquisitionmethods', views.DataAcquisitionMethodViewSet,
                        basename='dataacquisitionmethod')

urlpatterns = [
    path('api/', include(project_router.urls)),
    path('api/', include(experiment_router.urls)),
    path('api/', include(session_router.urls)),
]



import datetime
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('projects', '0003_project_project_manager_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='DataAcquisitionMethodType',
            fields=[
                ('short_name', models.CharField(editable=False, max_length=32, primary_key=True, serialize=False, verbose_name='Short name')),
                ('name', models.CharField(max_length=255, verbose_name='Name')),
            ],
            options={
                'verbose_name': 'Data acquisition method type',
                'verbose_name_plural': 'Data acquisition method types',
            },
        ),
        migrations.CreateModel(
            name='Experiment',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=255, verbose_name='Title')),
                ('description', models.TextField(blank=True, default='', verbose_name='Description')),
                ('order', models.PositiveSmallIntegerField(verbose_name='Order')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='experiments', to='projects.project', verbose_name='Project')),
            ],
            options={
                'verbose_name': 'Experiment',
                'verbose_name_plural': 'Experiments',
            },
        ),
        migrations.CreateModel(
            name='Session',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=255, verbose_name='Title')),
                ('order', models.PositiveSmallIntegerField(verbose_name='Order')),
                ('setup_time', models.DurationField(default=datetime.timedelta(0), verbose_name='Setup time')),
                ('duration', models.DurationField(verbose_name='Duration')),
                ('teardown_time', models.DurationField(default=datetime.timedelta(0), verbose_name='Teardown time')),
                ('experiment', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sessions', to='project_design.experiment', verbose_name='Experiment')),
            ],
            options={
                'verbose_name': 'Session',
                'verbose_name_plural': 'Sessions',
            },
        ),
        migrations.CreateModel(
            name='DataAcquisitionMethod',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('order', models.PositiveSmallIntegerField(verbose_name='Order')),
                ('stimulus', models.CharField(choices=[('auditive', 'Auditive'), ('gustatory', 'Gustatory'), ('haptic', 'Haptic'), ('olfactory', 'Olfactory'), ('visual', 'Visual')], max_length=32, null=True, verbose_name='Stimulus')),
                ('location', models.CharField(max_length=126, verbose_name='Location')),
                ('session', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='data_acquisition_methods', to='project_design.session', verbose_name='Session')),
                ('type', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='project_design.dataacquisitionmethodtype', verbose_name='Type')),
            ],
            options={
                'verbose_name': 'Data acquisition method',
                'verbose_name_plural': 'Data acquisition methods',
            },
        ),
    ]



from rest_framework import serializers

from huscy.project_design import models, services


class DataAcquisitionMethodSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.DataAcquisitionMethod
        fields = (
            'id',
            'location',
            'order',
            'session',
            'stimulus',
            'type',
        )
        read_only_fields = 'id', 'order', 'session'

    def get_fields(self):
        fields = super().get_fields()
        request = self.context.get('request')
        if request and request.method == 'PUT':
            fields['type'].read_only = True
        return fields

    def create(self, validated_data):
        return services.create_data_acquisition_method(**validated_data)

    def update(self, data_acquisition_method, validated_data):
        return services.update_data_acquisition_method(data_acquisition_method, **validated_data)


class SessionSerializer(serializers.ModelSerializer):
    data_acquisition_methods = DataAcquisitionMethodSerializer(many=True, read_only=True)
    title = serializers.CharField(required=False)

    class Meta:
        model = models.Session
        fields = (
            'id',
            'data_acquisition_methods',
            'duration',
            'experiment',
            'order',
            'title',
        )
        read_only_fields = 'id', 'experiment', 'order'

    def create(self, validated_data):
        return services.create_session(**validated_data)

    def update(self, session, validated_data):
        return services.update_session(session, **validated_data)


class ExperimentSerializer(serializers.ModelSerializer):
    sessions = SessionSerializer(many=True, read_only=True)
    title = serializers.CharField(required=False)

    class Meta:
        model = models.Experiment
        fields = (
            'id',
            'description',
            'order',
            'project',
            'sessions',
            'title',
        )
        read_only_fields = 'id', 'order', 'project',

    def create(self, validated_data):
        return services.create_experiment(**validated_data)

    def update(self, experiment, validated_data):
        return services.update_experiment(experiment, **validated_data)



from django.shortcuts import get_object_or_404
from rest_framework import mixins, viewsets
from rest_framework.permissions import IsAuthenticated

from huscy.project_design import models, serializer, services
from huscy.projects.models import Project


class ExperimentViewSet(mixins.CreateModelMixin, mixins.DestroyModelMixin, mixins.ListModelMixin,
                        mixins.UpdateModelMixin, viewsets.GenericViewSet):
    serializer_class = serializer.ExperimentSerializer
    permission_classes = (IsAuthenticated, )

    def initial(self, request, *args, **kwargs):
        self.project = get_object_or_404(Project, pk=self.kwargs['project_pk'])
        super().initial(request, *args, **kwargs)

    def get_queryset(self):
        return services.get_experiments(self.project)

    def perform_create(self, serializer):
        serializer.save(project=self.project)


class SessionViewSet(mixins.CreateModelMixin, mixins.DestroyModelMixin, mixins.UpdateModelMixin,
                     viewsets.GenericViewSet):
    queryset = models.Session.objects.all()
    permission_classes = (IsAuthenticated, )
    serializer_class = serializer.SessionSerializer

    def initial(self, request, *args, **kwargs):
        self.experiment = get_object_or_404(
            models.Experiment,
            pk=self.kwargs['experiment_pk'],
            project=self.kwargs['project_pk']
        )
        super().initial(request, *args, **kwargs)

    def perform_create(self, serializer):
        serializer.save(experiment=self.experiment)


class DataAcquisitionMethodViewSet(mixins.CreateModelMixin, mixins.DestroyModelMixin,
                                   mixins.UpdateModelMixin, viewsets.GenericViewSet):
    queryset = models.DataAcquisitionMethod.objects.all()
    serializer_class = serializer.DataAcquisitionMethodSerializer
    permission_classes = (IsAuthenticated, )

    def initial(self, request, *args, **kwargs):
        self.session = get_object_or_404(
            models.Session,
            experiment=self.kwargs['experiment_pk'],
            pk=self.kwargs['session_pk'],
            experiment__project=self.kwargs['project_pk']
        )
        super().initial(request, *args, **kwargs)

    def perform_create(self, serializer):
        serializer.save(session=self.session)
